{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ImportQualifiedPost #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE ViewPatterns #-}

module Main (main) where

import Control.Applicative (Alternative ((<|>)), Applicative (liftA2))
import Control.Monad.Random (evalRand)
import Data.Bifunctor (Bifunctor (bimap, first))
import Data.List (sortBy, union)
import Data.Map qualified as M
import Data.Maybe (isNothing)
import Data.Ord (Down (Down), comparing)
import Data.Set qualified as S
import Data.Text qualified as T
import System.Random (RandomGen, newStdGen)
import System.Random.Shuffle (shuffleM)

parseGraph :: [T.Text] -> M.Map T.Text [T.Text]
parseGraph =
  (M.fromListWith union . concatMap (liftA2 fmap (flip (,) . pure . fst) snd) . M.assocs >>= M.unionWith union)
    . M.fromList
    . fmap parseLine
  where
    parseLine (T.splitOn ": " -> [name, T.words -> nodes]) = (name, nodes)
    parseLine _ = error "parseLine error"

bfsWalk :: ([T.Text] -> [T.Text]) -> T.Text -> M.Map T.Text [T.Text] -> M.Map (T.Text, T.Text) Integer
bfsWalk shuffleMapper start graph = go [start] S.empty
  where
    go (node : xs) visited = foldl (flip (M.alter (fmap (+ 1) . (<|> Just 0)))) next (fmap (node,) neighbors)
      where
        neighbors = shuffleMapper $ filter (isNothing . flip S.lookupIndex visited) (graph M.! node)
        next = go (xs ++ neighbors) (foldl (flip S.insert) visited neighbors)
    go [] _ = M.empty

shuffleList :: (RandomGen g) => g -> [a] -> [a]
shuffleList = flip (evalRand . shuffleM)

genFlipped :: [(T.Text, T.Text)] -> [(T.Text, T.Text)]
genFlipped = (++) =<< fmap (uncurry (flip (,)))

nodeCnt :: M.Map (T.Text, T.Text) Integer -> Integer
nodeCnt = M.foldl (+) 0

main :: IO ()
main =
  do
    rng <- newStdGen
    graph <- parseGraph . fmap T.pack . lines <$> getContents

    let walkGraph = flip (bfsWalk (shuffleList rng))
    let connectingEdges =
          take 3
            . sortBy (comparing (Down . snd))
            . M.toList
            . M.fromListWith (+)
            . fmap (first ((,) <$> uncurry min <*> uncurry max))
            . M.toList
            . M.unionsWith (+)
            $ walkGraph graph
              <$> M.keys graph

    let edgesToDrop = genFlipped $ fmap fst connectingEdges
    let bisectedGraph = foldl (flip (\i -> M.adjust (filter (snd i /=)) (fst i))) graph edgesToDrop

    print $ length . M.keysSet $ graph
    let parts = bimap (nodeCnt . walkGraph bisectedGraph) (nodeCnt . walkGraph bisectedGraph) (head edgesToDrop)
    print parts
    print $ uncurry (*) parts